;;; xref.lisp -- a cross-reference facility for CMUCL
;;;
;;; Author: Eric Marsden <emarsden@laas.fr>
;;; Time-stamp: <2002-11-29 emarsden>
;;
;;
;; The cross-reference facility provides the ability to discover
;; information such as which functions call which other functions and
;; what global variables are used. The cross-referencer maintains a
;; database of cross-reference information which can be queried by the
;; user to provide answers to questions like "who calls function foo?"
;;
;; The cross-referencing facility is activated by enabling the
;; variable C::*RECORD-XREF-INFO*. 
;;
;;  - the list of functions that call a given function, either
;;    directly or indirectly (via its function-object).
;;
;;  - the list of functions that read a global variable (ie a dynamic
;;    variable or a constant variable -- something declared with
;;    DEFVAR or DEFPARAMETER or DEFCONSTANT).
;;
;;  - the list of functions that bind a global variable
;;
;;  - the list of functions that modify a global variable
;;
;;
;; Missing functionality:
;;
;;   - macros have strange names; you get a name of the form "DEFMACRO FOOBAR"
;;
;;   - CLOS code isn't handled very well; methods have names of the form
;;     "(defmethod foobar (t))"
;;
;;   - there is no mechanism to save cross-reference information in a
;;     FASL file, so you need to recompile all files in the current image
;;
;;
;; The cross-reference facility is implemented by walking the IR1
;; representation that is generated by CMUCL when compiling (for both
;; native and byte-compiled code).
;;
;; FIXME maybe add macros EXT:WITH-XREF and an :xref option to
;; COMPILE-FILE, as per ACL.
;;
;;
;;
;; http://www.franz.com/support/documentation/6.2/doc/cross-reference.htm
;;
;; In Lispworks, there appear to be similar mechanisms inside the
;; compiler, but it isn't clear how the information is made available
;; to the user.


(in-package :compiler)
(export '(who-calls init-xref-database))


;; functions that call a globally-defined function
(defvar *who-calls* (make-hash-table :test #'eq))

;; functions that reference a global variable
(defvar *who-uses* (make-hash-table :test #'eq))

;; functions that bind a global variable
(defvar *who-binds* (make-hash-table :test #'eq))

;; functions that set a global variable
(defvar *who-sets* (make-hash-table :test #'eq))

;; you can print these conveniently with code like
;; (maphash (lambda (k v) (format t "~S <-~{ ~S~}~%" k v)) c::*who-sets*)


(defmacro register-xref (source target database)
  `(if (gethash ,target ,database)
       (pushnew ,source (gethash ,target ,database) :test 'eq)
       (setf (gethash ,target ,database) (list ,source))))

;; INIT-XREF-DATABASE -- interface
;;
(defun init-xref-database ()
  "Reinitialize the cross-reference database."
  (setf *who-calls* (make-hash-table :test #'eq))
  (setf *who-uses* (make-hash-table :test #'eq))
  (setf *who-binds* (make-hash-table :test #'eq))
  (setf *who-sets* (make-hash-table :test #'eq))
  (values))

;; WHO-CALLS -- interface
;;
(defun who-calls (function)
  (gethash function *who-calls*))


;; RECORD-NODE-XREFS -- internal
;;
;; TOPLEVEL-NAME is an indication of the name of the COMPONENT that
;; contains this node, or NIL if it was really "Top-Level Form". This
;; is a hack, and it leads to duplication of functions named as
;; strings (when the call occurs within a LET, for example) and named
;; as symbols. It would be better to fix this so that for a call that
;; isn't at the topmost level, it manages to find the real function
;; name by following CONTs (or something).
(defun record-node-xrefs (node toplevel-name)
  (declare (type node node))
  (typecase node
    (ref
     (let* ((leaf (ref-leaf node))
            (lexenv (ref-lexenv node))
            (lambda (lexenv-lambda lexenv))
            (caller (and lambda (lambda-name lambda))))

       (unless caller (setf caller toplevel-name))

       (typecase leaf
         ;; a reference to a LEAF of type GLOBAL-VAR
         (global-var
          (let ((called (global-var-name leaf)))
            ;; a reference to #'C::%SPECIAL-BIND means that we are
            ;; binding a special variable. The information on which
            ;; variable is being bound, and within which function, is
            ;; available in the ref's LEXENV object.
            (cond ((eq called 'c::%special-bind)
                   (let ((caller (caar (lexenv-blocks lexenv)))
                         (called (caar (lexenv-variables lexenv))))
                     (register-xref caller called *who-binds*)))
                  ;; we're not interested in lexical environments that have no name
                  ((not caller)
                   (format t "XREF: ignoring call to ~a in ~a because nil lexical environment~%"
                           called lambda))
                  ;; we're not interested in lexical environments
                  ;; named "Top-Level Form".
                  ((and (stringp caller)
                        (string= "Top-Level Form" caller))
                   t)
                  ((not called)
                   (format t "XREF: ignoring call from ~a in ~a because caller ~a unnamed~%"
                           caller lambda))
                  ((eq :global-function (global-var-kind leaf))
                   (register-xref caller called *who-calls*))
                  ((eq :special (global-var-kind leaf))
                   (register-xref caller called *who-uses*)))))
         ;; a reference to a LEAF of type CONSTANT
         (constant
          (let ((called (constant-name leaf)))
            (unless caller (setf caller toplevel-name))
            (and called
                 (not (eq called t))    ; ignore trivial variables
                 caller
                 (not (and (stringp caller) (string= "Top-Level Form" caller)))
                 (register-xref caller called *who-uses*)))))))

    ;; a variable is being set
    (cset
     (let* ((variable (set-var node))
            (lexenv (set-lexenv node)))
       (and (global-var-p variable)
            (eq :special (global-var-kind variable))
            (let* ((lblock (first (lexenv-blocks lexenv)))
                   (user (or (and lblock (car lblock)) toplevel-name))
                   (used (global-var-name variable)))
              (and user used (register-xref user used *who-sets*))))))

    ;; nodes of type BIND are used to bind symbols to LAMBDA objects
    ;; (including for macros), but apparently not for bindings of
    ;; variables.
    (bind
     t)))

;; RECORD-COMPONENT-XREFS -- internal
;;
(defun record-component-xrefs (component)
  (declare (type component component))
  (do ((block (block-next (component-head component)) (block-next block)))
      ((null (block-next block)))
    (let ((fun (block-home-lambda block))
          ;; (name (find-component-name component))
          (name (component-name component))
          (this-cont (block-start block))
          (last (block-last block)))
      (unless (eq :deleted (functional-kind fun))
        (loop
         (let ((node (continuation-next this-cont)))
           (record-node-xrefs node name)
           (let ((cont (node-cont node)))
             (when (eq node last) (return))
             (setq this-cont cont))))))))


;; EOF
